@@ -2,7 +2,6 @@ import os, tqdm
 import numpy as np
 import matplotlib.pyplot as plt
 from astropy.io import fits
-from muchbettermoments import quadratic_2d
 from astropy.wcs import WCS, NoConvergence
 from astropy.table import Table
 from astropy.nddata import Cutout2D
@@ -15,25 +14,36 @@ import urllib
 from .mast import tic_by_contamination
 
 
-def load_pointing_model(sector, camera, chip):
-    """ Loads in pointing model from website.
+def check_pointing(sector, camera, chip, path=None):
+    """ Checks to see if a pointing model exists locally already.
     """
-    user_agent = 'eleanor 0.1.6'
-    values = {'name': 'eleanor',
-              'language': 'Python' }
-    headers = {'User-Agent': user_agent}
+    # Tries to create a pointing model directory
+    if path == None:
+        pm_dir = '.'
+    else:
+        pm_dir = path
 
-    data = urllib.parse.urlencode(values)
-    data = data.encode('ascii')
+    searches = [
+        's{0:04d}-{1}-{2}_tess_v2_pm.txt'.format(sector, camera, chip),
+        'pointingModel_{0:04d}_{1}-{2}.txt'.format(sector, camera, chip),
+    ]
 
-    guide_link = 'https://users.flatironinstitute.org/dforeman/public_www/tess/postcards_test/s{0:04d}/pointing_model/pointingModel_{0:04d}_{1}-{2}.txt'.format(sector, camera, chip)
+    # Checks a directory of pointing models, if it exists
+    # Returns the pointing model if it's in the pointing model directory
+    for search in searches:
+        if not os.path.isdir(pm_dir):
+            continue
+        pm_downloaded = os.listdir(pm_dir)
+        pm = [i for i in pm_downloaded if search in i]
+        if len(pm) > 0:
+            return Table.read(os.path.join(pm_dir, pm[0]), format="ascii.basic")
+    warnings.warn("couldn't find pointing model")
 
-    req = urllib.request.Request(guide_link, data, headers)
-    with urllib.request.urlopen(req) as response:
-        pointing = response.read().decode('utf-8')
 
-    pointing = Table.read(pointing, format='ascii.basic') # guide to postcard locations
-    return pointing
+def load_pointing_model(pm_dir, sector, camera, chip):
+    """ Loads in pointing model.
+    """
+    return check_pointing(sector, camera, chip, path=pm_dir)
 
 
 def use_pointing_model(coords, pointing_model):
@@ -75,7 +85,7 @@ def pm_quality(time, sector, camera, chip, pm=None):
         cent_x,cent_y = [], []
 
         if pm is None:
-            pm = load_pointing_model(sector, camera, chip)
+            pm = load_pointing_model(pm_dir, sector, camera, chip)
 
         # Applies centroids
         for i in range(len(pm)):
@@ -146,6 +156,88 @@ def set_quality_flags(ffi_start, ffi_stop, shortCad_fn, sector, camera, chip,
 
     return flags+pm_flags
 
+def centroid_quadratic(data, mask=None):
+    """Computes the quadratic estimate of the centroid in a 2d-array.
+
+    This method will fit a simple 2D second-order polynomial
+    $P(x, y) = a + bx + cy + dx^2 + exy + fy^2$
+    to the 3x3 patch of pixels centered on the brightest pixel within
+    the image.  This function approximates the core of the Point
+    Spread Function (PSF) using a bivariate quadratic function, and returns
+    the maximum (x, y) coordinate of the function using linear algebra.
+
+    For the motivation and the details around this technique, please refer
+    to Vakili, M., & Hogg, D. W. 2016, ArXiv, 1610.05873.
+
+    Caveat: if the brightest pixel falls on the edge of the data array, the fit
+    will tend to fail or be inaccurate.
+
+    As used in the Lightkurve package of Barentsen et al.
+
+    Parameters
+    ----------
+    data : 2D array
+        The 2D input array representing the pixel values of the image.
+    mask : array_like (bool), optional
+        A boolean mask, with the same shape as `data`, where a **True** value
+        indicates the corresponding element of data is masked.
+
+    Returns
+    -------
+    column, row : tuple
+        The coordinates of the centroid in column and row.  If the fit failed,
+        then (NaN, NaN) will be returned.
+    """
+    # Step 1: identify the patch of 3x3 pixels (z_)
+    # that is centered on the brightest pixel (xx, yy)
+    if mask is not None:
+        data = data * mask
+    arg_data_max = np.nanargmax(data)
+    yy, xx = np.unravel_index(arg_data_max, data.shape)
+    # Make sure the 3x3 patch does not leave the TPF bounds
+    if yy < 1:
+        yy = 1
+    if xx < 1:
+        xx = 1
+    if yy > (data.shape[0] - 2):
+        yy = data.shape[0] - 2
+    if xx > (data.shape[1] - 2):
+        xx = data.shape[1] - 2
+
+    z_ = data[yy-1:yy+2, xx-1:xx+2]
+
+    # Next, we will fit the coefficients of the bivariate quadratic with the
+    # help of a design matrix (A) as defined by Eqn 20 in Vakili & Hogg
+    # (arxiv:1610.05873). The design matrix contains a
+    # column of ones followed by pixel coordinates: x, y, x**2, xy, y**2.
+    A = np.array([[1, -1, -1, 1,  1, 1],
+                  [1,  0, -1, 0,  0, 1],
+                  [1,  1, -1, 1, -1, 1],
+                  [1, -1,  0, 1,  0, 0],
+                  [1,  0,  0, 0,  0, 0],
+                  [1,  1,  0, 1,  0, 0],
+                  [1, -1,  1, 1, -1, 1],
+                  [1,  0,  1, 0,  0, 1],
+                  [1,  1,  1, 1,  1, 1]])
+    # We also pre-compute $(A^t A)^-1 A^t$, cf. Eqn 21 in Vakili & Hogg.
+    At = A.transpose()
+    # In Python 3 this can become `Aprime = np.linalg.inv(At @ A) @ At`
+    Aprime = np.matmul(np.linalg.inv(np.matmul(At, A)), At)
+
+    # Step 2: fit the polynomial $P = a + bx + cy + dx^2 + exy + fy^2$
+    # following Equation 21 in Vakili & Hogg.
+    # In Python 3 this can become `Aprime @ z_.flatten()`
+    a, b, c, d, e, f = np.matmul(Aprime, z_.flatten())
+
+    # Step 3: analytically find the function maximum,
+    # following https://en.wikipedia.org/wiki/Quadratic_function
+    det = 4 * d * f - e ** 2
+    if abs(det) < 1e-6:
+        return np.nan, np.nan  # No solution
+    xm = - (2 * f * b - c * e) / det
+    ym = - (2 * d * c - b * e) / det
+    return xx + xm, yy + ym
+
 
 class ffi:
     """This class allows the user to download all full-frame images for a given sector,
@@ -339,7 +431,7 @@ class ffi:
 
 
         def isolated_center(x, y, image):
-            """Finds the centroid of each isolated source with quadratic_2d.
+            """Finds the centroid of each isolated source with centroid_quadratic.
 
             Parameters
             ----------
@@ -365,7 +457,7 @@ class ffi:
                  if x[i] > 0. and y[i] > 0.:
                      tpf = Cutout2D(image, position=(x[i], y[i]), size=(7,7), mode='partial')
                      origin = tpf.origin_original
-                     cen = quadratic_2d(tpf.data - np.nanmedian(tpf.data))
+                     cen = centroid_quadratic(tpf.data - np.nanmedian(tpf.data))
                      cenx.append(cen[0]+origin[0]); ceny.append(cen[1]+origin[1])
                      good.append(i)
             cenx, ceny = np.array(cenx), np.array(ceny)
@@ -377,10 +469,11 @@ class ffi:
             new_coords = use_pointing_model(np.array(xy).T, pointing_model)
             return np.array(new_coords)
 
+        pm_fn = 's{0:04d}-{1}-{2}_tess_v2_pm.txt'.format(self.sector, self.camera, self.chip)
 
-        pm_fn = 'pointingModel_{0:04d}_{1}-{2}.txt'.format(self.sector, self.camera, self.chip)
+        eleanorpath = os.path.join(os.path.expanduser('~'), '.eleanor')
 
-        qf = np.loadtxt('https://archipelago.uchicago.edu/tess_postcards/metadata/s{0:04d}/quality_s{0:04d}.txt'.format(self.sector,
+        qf = np.loadtxt(eleanorpath + '/metadata/s{0:04d}/quality_s{0:04d}.txt'.format(self.sector,
                                                                                                                         self.sector))
 
         if out_dir is not None:
@@ -389,9 +482,15 @@ class ffi:
         with open(pm_fn, 'w') as tf:
             tf.write('0 1 2 3 4 5 6 7 8\n')
 
-        hdu = fits.open(self.local_paths[0])
-        hdr = hdu[1].header
-        pos = [hdr['CRVAL1'], hdr['CRVAL2']]
+        pos = None
+        for fn in self.local_paths:
+            with fits.open(fn) as hdu:
+                hdr = hdu[1].header
+            if 'CRVAL1' not in hdr or 'CRVAL2' not in hdr:
+                continue
+            pos = [hdr['CRVAL1'], hdr['CRVAL2']]
+        if pos is None:
+            raise ValueError("no WCS")
 
         r = 6.0*np.sqrt(1.2)
         contam = [0.0, 0.5]
@@ -400,9 +499,10 @@ class ffi:
         t  = tic_by_contamination(pos, r, contam, tmag_lim).group_by('contratio')
 
         for i, fn in enumerate(self.local_paths):
-            hdu = fits.open(fn)
-            hdr = hdu[1].header
-            
+            with fits.open(fn) as hdu:
+                hdr = hdu[1].header
+                data = hdu[1].data
+
             try:
                 xy = WCS(hdr).all_world2pix(t['ra'], t['dec'], 1)
 
@@ -412,15 +512,15 @@ class ffi:
                 iso = find_isolated(xy[0], xy[1])
                 if len(iso) > 0:
                     xy  = np.array([xy[0][iso], xy[1][iso]])
-                    cenx, ceny, good = isolated_center(xy[0], xy[1], hdu[1].data)
-                    
+                    cenx, ceny, good = isolated_center(xy[0], xy[1], data)
+
                     # Triple checks there are no nans; Nans make people sad
                     no_nans = np.where( (np.isnan(cenx)==False) & (np.isnan(ceny)==False))
                     pos_inferred = np.array( [cenx[no_nans], ceny[no_nans]] )
                     xy = np.array( [xy[0][no_nans], xy[1][no_nans]] )
 
                     solution = self.build_pointing_model(xy.T, pos_inferred.T)
-                    
+
                     xy = apply_pointing_model(xy.T, solution)
                     matrix = self.build_pointing_model(xy, pos_inferred.T, outlier_removal=True)
 
@@ -436,7 +536,7 @@ class ffi:
                     a   = np.zeros((3, 3), int)
                     np.fill_diagonal(a, 1)
                     sol = np.reshape(a, (9,))
-                    
+
             with open(pm_fn, 'a') as tf:
                 tf.write('{}\n'.format(' '.join(str(e) for e in sol) ) )
 
